import 'dart:async';
import 'dart:io';

import 'package:analyzer/dart/element/element.dart';
import 'package:restoque_build/src/filters/equals.dart';
import 'package:restoque_build/src/model/constructor.dart';
import 'package:restoque_build/src/model/field.dart';
import 'package:source_gen/source_gen.dart';
import 'package:path/path.dart' as p;
import '../annotations.dart';
import 'model/modelVisitor.dart';

/// Generate code from @RestoqueEntity analysis.
class ConstructorGenerator {
  static final _entityKeyChecker =
      const TypeChecker.fromRuntime(RestoquePropertyName);

  static String path =
      p.join(Directory.current.path, '../restoque_build/lib/src');

  static Future<String> generate(constructor _constructor) async {
    String dataDir = Directory.current.path + "/lib/database/";
    bool exists = await File(dataDir + "AppDatabase.dart").exists();

    if (!exists) {
      exists = await Directory(dataDir).exists();
      if (!exists) await Directory(dataDir).create();

      await File(path + "/toCopy/AppDatabase.txt")
          .copy(dataDir + "AppDatabase.dart");
    }

    String finalKey = _getEntityKeys(_constructor.element);
    if (finalKey.isEmpty) {
      print("@RestoqueEntityKey not found.");
    }

    String filterFuntions = await _getFilters(_constructor);
    String fromJson = await _getFromJsonFunction(_constructor);
    String toJson = await _getToJsonFunction(_constructor);
    String content = await File('$path/restoqueGenericDao.txt').readAsString();

    content = content
        .replaceAll("@element", _constructor.className)
        .replaceAll("@Key", "$finalKey")
        .replaceAll("@TableName", _constructor.tableName)
        .replaceAll("@FilterFunctions", filterFuntions)
        .replaceAll("@fromJson", fromJson)
        .replaceAll("@toJson", toJson);

    return content;
  }

  /// List all fields in .dart class
  static List<field> _getFields(constructor _constructor) {
    List<field> fields = List<field>();

    for (var f in (_constructor.element as ClassElement).fields) {
      String fieldName = f.name;
      String fieldAnnotationName = f.name;
      if (_entityKeyChecker.hasAnnotationOfExact(f)) {
        fieldAnnotationName = _entityKeyChecker
            .firstAnnotationOfExact(f)
            .getField('name')
            .toStringValue();
      }

      field _field = field(
          fieldName: fieldName,
          fieldAnnotationName: fieldAnnotationName,
          dartType: f.type);
      fields.add(_field);
    }

    return fields;
  }

  /// Get the primary key of the entity
  static String _getEntityKeys(Element element) {
    String finalKey = "";
    var visitor = ModelVisitor();
    element.visitChildren(visitor);
    visitor.metaData.forEach((k, y) {
      for (var meta in y) {
        if (meta.toString().contains("@RestoqueEntityKey")) {
          finalKey += "instance.$k.toString()+";
        }
      }
    });
    if (finalKey.length > 0)
      finalKey = finalKey.substring(0, finalKey.length - 1);
    return finalKey;
  }

  /// List of the all filter that must be created.
  static Future<String> _getFilters(constructor _constructor) async {
    List<String> filterList = [
      "equals",
      "notEquals",
      "greaterThan",
      "greaterThanOrEquals",
      "lessThan",
      "lessThanOrEquals"
    ];
    String _filterFunctions = "";
    List<field> fields = _getFields(_constructor);
    for (field _field in fields) {
      for (String _type in filterList) {
        _filterFunctions += await _getFindEqual(
            _constructor, _field.fieldName, _field.fieldAnnotationName, _type);
      }
    }
    filterList.clear();
    filterList.add("isNull");
    filterList.add("notNull");
    for (field _field in fields) {
      for (String _type in filterList) {
        _filterFunctions += await _getFindNull(
            _constructor, _field.fieldName, _field.fieldAnnotationName, _type);
      }
    }
    return _filterFunctions;
  }

  ///Return the functions of the filters
  static Future<String> _getFindEqual(constructor _constructor,
      String fieldname, String fieldAnnotation, String filterType) async {
    String content = Equals().equals;
    content = content
        .replaceAll("@element", _constructor.className)
        .replaceAll("@fieldAnnotation", fieldAnnotation)
        .replaceAll("@fieldType_",
            filterType[0].toUpperCase() + filterType.substring(1))
        .replaceAll("@fieldType", filterType)
        .replaceAll("@field", fieldname);
    return content;
  }

  /// Return the functions of the filters isNull and notNull
  static Future<String> _getFindNull(constructor _constructor, String fieldname,
      String fieldAnnotation, String filterType) async {
    String content = await File('$path/filters/isNull.txt').readAsString();
    content = content
        .replaceAll("@element", _constructor.className)
        .replaceAll("@fieldAnnotation", fieldAnnotation)
        .replaceAll("@fieldType_",
            filterType[0].toUpperCase() + filterType.substring(1))
        .replaceAll("@fieldType", filterType)
        .replaceAll("@field", fieldname);
    return content;
  }

  ///Return the function of FromJson
  static Future<String> _getFromJsonFunction(constructor _constructor) async {
    String content = await File('$path/json/fromJson.txt').readAsString();

    List<field> fields = await _getFields(_constructor);
    String functionBody = "";
    for (field _field in fields) {
      if (_field.dartType.displayName == "double" ||
          _field.dartType.displayName == "String" ||
          _field.dartType.displayName == "int" ||
          _field.dartType.displayName == "bool") {
        if (_field.dartType.displayName != "double") {
          functionBody +=
              " ${_field.fieldName}:json['${_field.fieldAnnotationName}'] as ${_field.dartType.displayName},";
        } else {
          functionBody +=
              " ${_field.fieldName}:(json['${_field.fieldAnnotationName}'] as  num ) ?.toDouble(),";
        }
      } else {
        if (_field.dartType.displayName.contains("List")) {
          String contentType =
              await File('$path/json/listType.txt').readAsString();
          String listType = _field.dartType.displayName
              .replaceAll("List<", "")
              .replaceAll(">", "");

          functionBody += " ${_field.fieldName}:" +
              contentType
                  .replaceAll("@fieldName", _field.fieldAnnotationName)
                  .replaceAll("@element", listType) +
              ",";
        } else {
          String sJson =
              " ${_field.fieldName}: json['@fieldName'] == null ? null: @element.fromJson(json['@fieldName'] as Map<String, dynamic>),";
          sJson = sJson
              .replaceAll("@fieldName", _field.fieldAnnotationName)
              .replaceAll("@element", _field.dartType.displayName);
          functionBody += sJson;
        }
      }
    }
    content = content
        .replaceAll("@element", _constructor.className)
        .replaceAll("@body", functionBody);
    return content;
  }

  ///Return the function of ToJson
  static Future<String> _getToJsonFunction(constructor _constructor) async {
    String content = await File('$path/json/toJson.txt').readAsString();
    List<field> fields = await _getFields(_constructor);
    String functionBody = "";
    for (field _field in fields) {
      functionBody +=
          "'${_field.fieldAnnotationName}': instance.${_field.fieldName},";
    }
    content = content
        .replaceAll("@element", _constructor.className)
        .replaceAll("@body", functionBody);
    return content;
  }
}
